#!/usr/bin/env python
# -*- coding: latin-1 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.

"""
How it works:
-------------

!! Beware of side effects. Use the -n parameter before uploading. !!
This script allows you to upload new SRR files to srrdb.com:
   srrdb.py first.file.srr second.release.srr
It also allows you to _batch_ upload additional files such .srs files:
   srrdb.py /path/to/a/directory/ C:\or\Windows\style
   srrdb.py . -n
Each directory must be structured in the following way:
   +---Release-Name/
   |       file.to.store.srs
   |
   +---Other.Release-Name/
   |   |
   |   +---Sample/
   |       unreconstructable.sample.avi.txt
   |...
It will walk the tree and tries to detect the release name based on bad
directory names.
SRS files are automatically put in a Sample/ folder. Other files do
not get a folder set.
See _SUPPORTED_FILES for the type of files that are detected for upload.

Download Python 2.7 from http://www.activestate.com/activepython/downloads

Install the dependencies that are necessary for this script:
  pypm.exe install poster
   or if you use the regular CPython version:
  easy_install.exe poster
http://pypi.python.org/pypi/setuptools

Version history:
0.1 (2011-10-17) http://www.mediafire.com/file/9bidbasgtg2nggr/srrdb_0.1.zip
    - Initial release available for public usage
0.2 (2011-11-24) 
    - recursively adds files (when srs is in Sample directory)
    - clean up .txt files

Author: Gfy <tsl@yninovg.pbz>
"""

from __future__ import unicode_literals
from poster.encode import multipart_encode
from poster.streaminghttp import register_openers
from urllib2 import HTTPCookieProcessor, ProxyHandler, Request
import urllib
import urllib2
import cookielib
import re
import os
import optparse
import sys
import time
import ConfigParser
import codecs

# supported extensions for files to add to a release
# checks config file and -e parameter later on
_SUPPORTED_FILES = (".srs", ".avi.txt", ".mkv.txt", ".srr") 

__version__ = "0.2"
_USER_AGENT = "Gfy's srrDB upload script version %s." % __version__

# some configuration options
_PROXY = False
_PROXY_TYPE = "http"
_PROXY_URL = "http://127.0.0.1:8008" # WebScarab

# srrdb.com credentials
_USERNAME = ""
_PASSWORD = ""
_URL = "http://www.srrdb.com/"

def fix_txt(filename):
    """ Cleans up .ext.txt files created by the DOS script. """
    if options.dry_run:
        print("Fixing '%s'." % filename)
        return
    data = codecs.open(filename, encoding="latin-1", mode="r").read()
    newdata = re.sub(".*Corruption", "Corruption", data)
    newdata = newdata.replace("\xff", ".") # cp850 \xa0
    newdata = re.sub("\r(?!\n)|(?<!\r)\n", "\r\n", newdata)
    if newdata != data:
        print("'%s' fixed." % filename)
        f = open(filename, "wb")
        f.write(newdata)
        f.close()
    else:
        print("Nothing done for %s!" % filename)
        
def addDicts(a, b):
    """ Joins two hashmaps """
    c = {}
    for item in a:
        c[item] = a[item]
    for item in b:
        c[item] = b[item]
    return c

class urlErrorDecorator(object):
    """ Decorator to share the same error handling code for the url calling
    stuff needed by a lot of functions.
    A nice intro to decorators:
    http://www.artima.com/weblogs/viewpost.jsp?thread=240808 
    http://wiki.python.org/moin/PythonDecoratorLibrary """
    def __init__(self, f):
        self.f = f

    def __call__(self, *args, **kwargs):
        try:
            return self.f(*args, **kwargs)
        except (urllib2.URLError, urllib2.HTTPError), e:
            if hasattr(e, "code"):
                print("We failed with error code - %s." % e.code)
            elif hasattr(e, 'reason'):
                print("The error object has the following 'reason' attribute:",
                      e.reason)
                print("This usually means the server doesn't exist, is down, "
                      "or you don't have an Internet connection.")

class Srrdb(object):
    """ Class that supports stuff from srrdb.com """
    def __init__(self, username, password):
        self.cj = cj = cookielib.CookieJar()
        
        # set up proxy to check sent requests with WebScarab
        if _PROXY:
            proxy_support = ProxyHandler({_PROXY_TYPE : _PROXY_URL})
            self.opener = urllib2.build_opener(proxy_support, 
                                HTTPCookieProcessor(cj))
        else:
            self.opener = urllib2.build_opener(HTTPCookieProcessor(cj))
        urllib2.install_opener(self.opener)
        
        self.baseurl = _URL
        self.username = username
        self.body = {
                'username' : username,
                'password' : password,
                'login': 'Login!'
                }
        self.headers = { 
                'Referer' : self.baseurl,
                'User-Agent' : _USER_AGENT,
                }
        self._login(self)

    @urlErrorDecorator
    def _login(self):
        """ Authenticate. """  
        data = urllib.urlencode(self.body)
        request = urllib2.Request(
            self.baseurl + "login.php", data, self.headers)
        HTMLSource = self.opener.open(request).read()
        res = re.findall(".*Welcome, .+%s.+" % self.username, HTMLSource)
        if len(res):
            print("Authentication successful.")
            match = re.match(".*profile\.php\?uid=(\d+).*", res[0])
            self.user_id = match.group(1)
            print("%s has user id %s." % (self.username, self.user_id))
        else:
            print("Authentication unsuccessful.") # or the site has changed

    @urlErrorDecorator
    def add_file(self, release, filename, folder):
        """ release: the release name
        filename: path where to find the file on HD
        folder: the folder to the SRR e.g. "Sample" without the /
        """
        # http://stackoverflow.com/questions/680305/
        # using-multipartposthandler-to-post-form-data-with-python
        # Register the streaming http handlers with urllib2
        opener = register_openers()
        # https://bitbucket.org/chrisatlee/poster/issue/7/
        # multipart-form-post-doesnt-work-with
        if _PROXY:
            opener.add_handler(ProxyHandler({_PROXY_TYPE : _PROXY_URL}))
                        
        # Start the multipart/form-data encoding of the file "DSC0001.jpg"
        # "image1" is the name of the parameter, which is normally set
        # via the "name" parameter of the HTML <input> tag.
        
        # headers contains the necessary Content-Type and Content-Length
        # datagen is a generator object that yields the encoded parameters
        datagen, headers = multipart_encode({
                "folder" : folder,
                "MAX_FILE_SIZE" : 26214400,
                "file": open(filename, "rb"),
                "add": "Add",})
        headers = addDicts(self.headers, headers)
        url = self.baseurl + "addfile.php?release=" + release
        request = Request(url, datagen, headers)
        opener.add_handler(HTTPCookieProcessor(self.cj))

        # Actually do the request, and get the response
        handle = urllib2.urlopen(request)
        html_source = handle.read()
        
        fn = folder + "/" + os.path.basename(filename)
        if len(re.findall(".*" + fn + ".*", html_source)):
            print("'%s' successfully uploaded." % fn)
            # also gives this result if it was already there in the first place
        elif len(re.findall(".*Release does not exist.*", html_source)):
            print("!!! '%s': no such release." % release)
        elif len(re.findall(".*an error occurred while adding the file.*", 
                            html_source)):  
            print("!!! '%s': file already added." % fn)
        else:
            print(html_source)
   
    @urlErrorDecorator
    def add_release(self, srr_file):
        """ srr_file: the srr file to upload  """
        opener = register_openers()
        if _PROXY:
            opener.add_handler(ProxyHandler({_PROXY_TYPE : _PROXY_URL}))
        datagen, headers = multipart_encode({
                "MAX_FILE_SIZE" : 26214400,
                "file": open(srr_file, "rb"),
                "upload": "Upload",})
        headers = addDicts(self.headers, headers)
        
        url = self.baseurl + "upload.php"
        request = Request(url, datagen, headers)
        opener.add_handler(HTTPCookieProcessor(self.cj))

        # Actually do the request, and get the response
        handle = urllib2.urlopen(request)
        html_source = handle.read()
        
        # <span style="font-weight: bold; color: #008000;"><a href="details.php
        # ?release=RELNAME" style="font-weight: bold; color: #008000;">RELNAME
        # </a> was added.</span><br />
        
        # <span style="font-weight: bold; color: #FF0000;"><a href="details.php
        # ?release=RELNAME" style="font-weight: bold; color: #FF0000;">RELNAME
        # </a> already exists.</span><br />
        
        if len(re.findall(".*</a> was added\.</span><br />.*", html_source)):
            print("'%s' was added." % srr_file)
        elif len(re.findall(".*</a> already exists.*", html_source)):
            print("!!! '%s' already exists." % srr_file)
        else:
            print(html_source)
    
def list_releases():
    """ Prints out a list of all releases. This includes the upload date and
    the availability of nfo and srs. """
    for number in range(1, get_nb_pages() + 1):
        for release in get_releases(number):
            print(release)
        if number % 100 == 0:
            time.sleep(5) # so we won't be hammering the site too much
    
@urlErrorDecorator
def get_releases(page_number):
    """ Returns a list with release names. 
    page_number: the number of the browse.php page """
    request = Request(url=_URL + "browse.php?page=%d" % page_number,
                      headers={'User-Agent' : _USER_AGENT})
    response = urllib2.urlopen(request).read()
    return re.findall('.*<a title="(?P<release>.+)" href="details.php\?'
                      'release=(?P=release)">.*</a></div></td>\n'
                      '\t+<td>(?P<adddate>\d{4}-\d{2}-\d{2})</td>\n'
                      '\t+<td class="(?P<nfo>low|high)"></td>\n'
                      '\t+<td class="(?P<sample>low|medium|high)"></td>.*',
                      response, re.MULTILINE)
              
@urlErrorDecorator
def get_nb_pages():
    """ Get the amount of pages from the browse.php page. """  
    request = urllib2.Request(url="http://www.srrdb.com/browse.php")
    response = urllib2.urlopen(request).read()
    match = re.search(".*<a href=\"browse.php\?page="
                     "(\d+)\">Last</a><br />.*", response)
    print("Number of pages: %s" % match.group(1))
    return int(match.group(1))

def guess_releasename(path):
    (head, tail) = os.path.split(path)
    if re.match("(vob)?(sample|subs|proof|covers?)", tail, re.IGNORECASE):
        return guess_releasename(head)
    else:
        return tail
        
def read_config():
    """The configuration file is in the same directory as the application."""
    folder = os.path.dirname(os.path.realpath(sys.argv[0]))
    config = ConfigParser.ConfigParser()
    global _SUPPORTED_FILES, _USERNAME, _PASSWORD, _URL,  \
            _PROXY, _PROXY_TYPE, _PROXY_URL
    cfile = os.path.join(folder, 'srrdb.cfg')
    try:
        config.readfp(open(cfile))
        _USERNAME = config.get("login", "username")
        _PASSWORD = config.get("login", "password")
        _URL = config.get("site", "url")
        _SUPPORTED_FILES = config.get("extensions", "ext").split(",")
        if config.getboolean("site", "proxy_enabled"):
            _PROXY = True
            _PROXY_TYPE = config.get("site", "proxy_type")
            _PROXY_URL = config.get("site", "proxy_url")
    except IOError, e:
        # create config file
        config.add_section("login")
        config.set("login", "username", "username")
        config.set("login", "password", "password")
        
        config.add_section("site")
        config.set("site", "url", _URL)
        config.set("site", "proxy_type", _PROXY_TYPE)
        config.set("site", "proxy_url", _PROXY_URL)
        config.set("site", "proxy_enabled", _PROXY)
        
        config.add_section("extensions")
        config.set("extensions", "ext", ",".join(_SUPPORTED_FILES))
        
        config.write(open(cfile, 'w'))
    except ConfigParser.NoOptionError, e:
        print(e)

def main(options, args):
    read_config()
    
    # overwrite config if an other parameter is given
    global _SUPPORTED_FILES, _USERNAME, _PASSWORD
    if options.extensions:
        _SUPPORTED_FILES = options.extensions.split(',')
    if options.login:
        _USERNAME = options.login
    if options.password:
        _PASSWORD = options.password
        
    print("Site: %s" % _URL)
    if _PROXY:
        print("Proxy: %s" % _PROXY_URL)
             
    if options.rellist:
        if options.dry_run:
            print("A dry run for getting the release names is not possible.")
        list_releases()
        sys.exit()
    
    if not options.dry_run:
        try:
            s = Srrdb(_USERNAME, _PASSWORD)
        except:
            print("Uploading SRR files as anonymous.")
    else:
        print("Script running as a dry run: no site lookups are done.")

    def process_file(path, file):
        if file[-4:] in _SUPPORTED_FILES or file[-8:] in _SUPPORTED_FILES:
            relname = guess_releasename(path)
            print("Storing file '%s' with release" % file)
            print("             '%s'." % relname)
            if file[-4:] in (".srs", ".txt"):
                srr_dir = "Sample"                  
            if not options.dry_run:
                srr_dir = ""
                if file[-4:] in (".srs", ".txt"):
                    srr_dir = "Sample"
                os.chdir(path)
                s.add_file(s, relname, file, srr_dir)
                  
    for element in args:
        element = os.path.abspath(element)
        dirname = os.path.dirname(element)
        basename = os.path.basename(element)
        if os.path.isfile(element) and element[-4:] == ".srr":
            if options.dry_run:
                print("Uploading '%s'." % element)
            else:
                # change current working dir 
                # (so no path info is send to the server
                os.chdir(dirname)
                s.add_release(s, basename)
        elif os.path.isdir(element):
            print("Processing files to upload (%s)" % element)
            # add srs, jpg, jpeg, png files to the actual release on site
            # only walk folders and add these files
            for dirpath, _dirnames, filenames in os.walk(element):
                for file in filenames:
                    if file[-4:] == ".txt" and options.fix_txt:
                        fix_txt(os.path.join(dirpath, file))
                    if file[-4:] == ".srr" and file[-4:] in _SUPPORTED_FILES:
                        if options.dry_run:
                            print("Uploading '%s'." % file)
                        else:
                            os.chdir(dirpath)
                            s.add_release(s, file)
                    else:
                        process_file(dirpath, file)
        else:
            print("WTF are you supplying me?")
         
if __name__ == '__main__':
    parser = optparse.OptionParser(
        usage="Usage: %prog [SRR file(s)] [directories] [options]'\n"
        "This tool will upload new SRR files or store additional files "
        "such as samples on srrdb.com.\n",
        version="%prog " + __version__) # --help, --version
    
    auth = optparse.OptionGroup(parser, "Authentication")
    auth.set_description("The credentials that are needed to login to "
                         "srrdb.com. SRR files will but uploaded anonymously "
                         "if no or no correct credentials are supplied.")
    parser.add_option_group(auth)
    
    auth.add_option("-l", "--login", help="supply username", 
                     metavar="USERNAME", default=_USERNAME, dest="login")
    auth.add_option("-p", "--password", help="supply password", 
                     metavar="PASSWORD", default=_PASSWORD, dest="password")
    
    parser.add_option("-n", "--dry-run", help="do no harm",
                      action="store_true", dest="dry_run", default=False)
    parser.add_option("-e", "--extensions", 
                      help="only add files with one of these extensions "
                      "to a release",
                      action="store", dest="extensions")
    parser.add_option("-t", "--fix-txt", help="fixes .txt files",
                      action="store_true", dest="fix_txt", default=False)
    parser.add_option("--rellist", help="list all releases",
                      action="store_true", dest="rellist", default=False) 
    
    # no arguments given
    if len(sys.argv) < 2:
        print(parser.format_help())
    else:       
        (options, args) = parser.parse_args()
        main(options, args)
    